# adapted from https://www.tensorflow.org/guide/keras/distributed_training

import logging
import os
from contextlib import nullcontext
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Optional, Union

import click
import mlflow
from mlflow_utils import (
    MLflowMetricsCallback,
    MlflowTimingCallback,
    TimingCallback,
    mlflow_log_sbatch_logs,
    mlflow_log_sbatch_scripts,
)

import tensorflow as tf
logger = logging.getLogger(__name__)


class TimingCallback(tf.keras.callbacks.Callback):
    def __init__(
        self,
        batch_size: int,
        log_freq: Union[None, str, int] = "batch",
        num_warmup_batches: int = 10,
        rank: int = 0,
    ):
        super().__init__()
        self.batch_size = batch_size
        self._chief_worker_only = True
        self.rank = rank
        self.num_warmup_batches = num_warmup_batches
        self.tracked_time = []
        if log_freq == "batch":
            self.log_every_epoch = False
            self.log_every_n_steps = 1
        elif isinstance(log_freq, int):
            self.log_every_epoch = False
            self.log_every_n_steps = log_freq
        else:
            raise ValueError(
                "Expected `log_freq` argument to be an integer or 'batch'. "
                f"Received: log_freq={log_freq} (of type {type(log_freq)})"
            )

        logger.info(str(self.__dict__))

    def on_train_batch_begin(self, batch, logs=None):
        if self.log_every_n_steps is None or logs is None:
            return

        current_iteration = int(self.model.optimizer.iterations.numpy())

        if (current_iteration + 1) % self.log_every_n_steps == 0 and self.rank == 0:
            self.batch_begin_time_train = perf_counter()

    def on_train_batch_end(self, batch, logs=None):
        if self.log_every_n_steps is None or logs is None:
            return

        current_iteration = int(self.model.optimizer.iterations.numpy())

        if current_iteration % self.log_every_n_steps == 0 and self.rank == 0:
            batch_end_time = perf_counter()

            time_elapsed_for_batch = batch_end_time - self.batch_begin_time_train

            logger.info(
                f"Current IPS: epoch_idx = {self.current_epoch}; batch_idx = {batch}; batch_time_elapsed = {time_elapsed_for_batch}; batch_images_per_sec = {self.batch_size / time_elapsed_for_batch}"
            )

            self.tracked_time.append(
                {
                    "type": "batch",
                    "epoch_idx": self.current_epoch,
                    "batch_idx": batch,
                    "begin_time": self.batch_begin_time_train,
                    "end_time": batch_end_time,
                    "batch_time_elapsed-train": time_elapsed_for_batch,
                    "batch_images_per_sec-train": self.batch_size
                    / time_elapsed_for_batch,
                    "rank": self.rank,
                    "step": current_iteration,
                }
            )

    def on_epoch_begin(self, epoch, logs=None):
        self.current_epoch = epoch

    def on_train_end(self, logs=None):
        if self.rank == 0:
            # print(self.tracked_time, flush = True)
            tracked_time_pd = pd.DataFrame.from_records(self.tracked_time)
            # print(tracked_time_pd.head(), flush = True)

            tracked_time_pd = tracked_time_pd[
                (tracked_time_pd["rank"] == 0)
                & (tracked_time_pd["step"] >= self.num_warmup_batches)
            ]
            logger.info(
                f"Average IPS = {tracked_time_pd['batch_images_per_sec-train'].mean()}(+-{tracked_time_pd['batch_images_per_sec-train'].std()})"
            )



def set_seed(seed: int = 5):
    import random
    import numpy as np
    
    if not isinstance(seed, int):
        raise ValueError(
            "Expected `seed` argument to be an integer. "
            f"Received: seed={seed} (of type {type(seed)})"
        )

    random.seed(seed)
    np.random.seed(seed)
    tf.random.set_seed(seed)



class NullStrategy:
    @staticmethod
    def scope():
        return nullcontext()

    @property
    def num_replicas_in_sync(self):
        return 1


class NullCommunication:
    @property
    def name(self):
        return "null"


@dataclass
class TrainingStrategy:
    """
    Determine and initialize the appropriate TensorFlow distribution strategy
    based on hardware configuration and environment.
    """

    device_type: str = field(init=False)
    num_nodes: int = field(init=False)
    num_gpus_per_node: int = field(init=False)
    strategy_type: str = field(init=False)
    communication_type: Union[
        str, tf.distribute.experimental.CommunicationImplementation
    ] = field(init=False)
    
    cross_device_communication_type: Union[
        str, tf.distribute.CrossDeviceOps
    ] = field(init=False)
    
    communication_options: Optional[tf.distribute.experimental.CommunicationOptions] = (
        field(default=None, init=False)
    )
    strategy: Union[
        tf.distribute.MultiWorkerMirroredStrategy,
        tf.distribute.MirroredStrategy,
        NullStrategy,
    ] = field(init=False)

    def __post_init__(self):
        self._gather_cluster_info()
        self.device_type = self._detect_device_type()
        self._select_strategy()
        self._log_strategy_params()

    def _gather_cluster_info(self) -> None:
        """Read environment variables related to cluster configuration."""
        self.num_nodes = int(os.environ.get("SLURM_NNODES", -1))
        self.num_gpus_per_node = int(os.environ.get("SLURM_GPUS_ON_NODE", -1))

    def _detect_device_type(self) -> str:
        """
        Determine the GPU type from available hardware.

        Returns:
            str: "NVIDIA", "AMD", or raises NotImplementedError if unknown.
        """
        gpus = tf.config.list_physical_devices("GPU")
        device_names = [
            tf.config.experimental.get_device_details(gpu).get("device_name", "")
            for gpu in gpus
        ]

        if not device_names:
            return "NONE"

        if all("NVIDIA" in name for name in device_names):
            return "NVIDIA"

        elif all("AMD" in name for name in device_names):
            return "AMD"

        raise NotImplementedError(f"Unsupported GPU types detected: {device_names}")

    def _select_strategy(self) -> None:
        """Select and initialize the appropriate distribution strategy."""
        if self.num_nodes == 1 and self.num_gpus_per_node == 1:
            self._use_null_strategy()
        elif self.num_nodes == 1:
            self._use_single_node_multi_gpu_strategy()
        elif self.num_nodes > 1:
            self._use_multi_node_strategy()
        else:
            raise NotImplementedError(
                f"Unsupported combination of num_nodes={self.num_nodes} and num_gpus_per_node={self.num_gpus_per_node}"
            )

    def _use_null_strategy(self) -> None:
        self.strategy_type = "null"
        self.communication_type = NullCommunication()
        self.cross_device_communication_type = NullCommunication()
        self.strategy = NullStrategy()

    def _use_single_node_multi_gpu_strategy(self) -> None:
        self.strategy_type = "MirroredStrategy"
        self.communication_type = NullCommunication()
        self.cross_device_communication_type =  self._get_cross_device_ops_implementation(self.device_type)
        self.strategy = tf.distribute.MirroredStrategy(cross_device_ops = self.cross_device_communication_type)
        

    def _use_multi_node_strategy(self) -> None:
        self.cross_device_communication_type = NullCommunication()

        self.communication_type = self._get_communication_implementation(
            self.device_type
        )
        self.communication_options = tf.distribute.experimental.CommunicationOptions(
            implementation=self.communication_type
        )

        self.strategy_type = "MultiWorkerMirroredStrategy"
        self.strategy = tf.distribute.MultiWorkerMirroredStrategy(
            communication_options=self.communication_options
        )
    
    def _get_cross_device_ops_implementation(self, device_type: str):
        """Map device type to appropriate communication implementation."""
        if device_type == "NVIDIA":
            return tf.distribute.NcclAllReduce()
        elif device_type == "AMD":
            return tf.distribute.ReductionToOneDevice()
        else:
            raise NotImplementedError(
                f"Unsupported device type for communication: {device_type}"
            )

    def _get_communication_implementation(self, device_type: str):
        """Map device type to appropriate communication implementation."""
        if device_type == "NVIDIA":
            return tf.distribute.experimental.CommunicationImplementation.NCCL
        elif device_type == "AMD":
            return tf.distribute.experimental.CommunicationImplementation.RING
        else:
            raise NotImplementedError(
                f"Unsupported device type for communication: {device_type}"
            )

    def _log_strategy_params(self) -> None:
        """Log key strategy configuration to MLflow."""
        logger.info(f"num_replicas_in_sync = {self.strategy.num_replicas_in_sync}")
        logger.info(f"strategy_type = {self.strategy_type}")
        logger.info(f"communication_type = {self.communication_type.name}")
        logger.info(f"cross_device_communication_type = {type(self.cross_device_communication_type)}")
        


@dataclass
class TrainingOptions:
    batch_size_per_device: int = 256
    num_epochs: int = 100
    lr: float = 1e-3
    model_type: str = "ResNet50"
    training_strategy: TrainingStrategy = field(init=False)
    momentum: float = 0.9
    weight_decay: float = 5e-4
    optimizer_name: str = "SGD"
    loss: str = "SparseCategoricalCrossentropy"
    seed: int = 228
    timing_log_freq: Union[int, str] = 10
    timing_warmup_batches: int = 10
    dataset_size: int = 20000
    global_batch_size: int = field(init=False)

    @classmethod
    def from_yaml(
        cls, cfg_path: Union[str, Path], cli_kwargs: Dict[Any, Any]
    ) -> "TrainingOptions":
        import yaml

        with open(cfg_path, "r") as file:
            data = yaml.safe_load(file)

        for key, item in cli_kwargs.items():
            logger.info(f"{key}: {item}")
            if item is not None:
                data[key] = item
        return cls(**data)

    def __post_init__(self):
        self.training_strategy = TrainingStrategy()
        self.distributed = self.training_strategy.strategy_type != "null"
        self.num_replicas_in_sync = (
            self.training_strategy.strategy.num_replicas_in_sync
            if hasattr(self.training_strategy.strategy, "num_replicas_in_sync")
            else 1
        )

        logger.info("Local batch size was provided, so scaling up gbs")
        self.global_batch_size = (
            int(self.batch_size_per_device * self.num_replicas_in_sync)
            if self.distributed
            else self.batch_size_per_device
        )

        self.steps_per_epoch = self.dataset_size // self.global_batch_size


class SYNTH_classifier:
    def __init__(self, opts: TrainingOptions):
        self.opts = opts

    def get_compiled_model(self):
        model = tf.keras.applications.ResNet50(weights=None)

        model.compile(
            optimizer=tf.keras.optimizers.SGD(
                learning_rate=self.opts.lr,
                momentum=self.opts.momentum,
                weight_decay=self.opts.weight_decay,
            ),
            loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
            metrics=[tf.keras.metrics.SparseCategoricalAccuracy()],
        )

        return model

    def prepare_dataset(self):
        
        @tf.function
        def gen_fn(_):
            image = tf.random.uniform([224, 224, 3])
            label = tf.random.uniform([], minval=0, maxval=999, dtype=tf.int64)
            return image, label

        dataset = tf.data.Dataset.range(10**9)  # # Infinite-like
        dataset = dataset.map(gen_fn, num_parallel_calls=tf.data.AUTOTUNE)
        dataset = dataset.batch(self.opts.global_batch_size, drop_remainder=True)
        dataset = dataset.prefetch(tf.data.AUTOTUNE)

        return dataset

    def train(self):
        
        self.train_dataset = self.prepare_dataset()
        logger.info(f"train_dataset: {type(self.train_dataset)}")
        
        # Define distributed strategy
        if self.opts.distributed:
            self.train_dataset = (
                self.opts.training_strategy.strategy.experimental_distribute_dataset(
                    self.train_dataset
                )
            )

        # Create a MirroredStrategy or MultiWorkerMirroredStrategy in case of distributed training, or just NullStrategy instead.
        with self.opts.training_strategy.strategy.scope():
            self.model = self.get_compiled_model()

        callbacks = self.configure_callbacks()
        logger.info(f"element_spec: {self.train_dataset.element_spec}")
        self.model.fit(
            x=self.train_dataset,
            epochs=self.opts.num_epochs,
            steps_per_epoch=self.opts.steps_per_epoch,
            callbacks=callbacks,
        )

    def configure_callbacks(self):
        callbacks = [
            TimingCallback(
                batch_size=self.opts.global_batch_size,
                log_freq=self.opts.timing_log_freq,
                rank=int(os.environ["RANK"]),
                num_warmup_batches=self.opts.timing_warmup_batches,
            ),
        ]

        return callbacks

    def evaluate(self, test_dataset):
        test_loss, test_acc = self.model.evaluate(test_dataset, verbose=0)
        return test_loss, test_acc

@click.group()
def cli():
    pass

@click.command(no_args_is_help=True)
@click.option("--run_cfg", type=click.Path(exists=True))
@click.option("--batch_size_per_device", type=int, default=None)
def train(
    run_cfg,
    batch_size_per_device,
):
    training_options = TrainingOptions.from_yaml(
        cfg_path=run_cfg,
        cli_kwargs=dict(
            batch_size_per_device=batch_size_per_device,
        ),
    )

    set_seed(training_options.seed)
    mnist_classifier = SYNTH_classifier(opts=training_options)
    mnist_classifier.train()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    logging.basicConfig(level=logging.INFO)  # Configure the root logger

    urllib3_logger = logging.getLogger("urllib3")
    urllib3_logger.setLevel(logging.WARNING)
    simple_parsing_logger = logging.getLogger("simple_parsing")
    simple_parsing_logger.setLevel(logging.INFO)
    
    cli()
